import asyncio
import logging
import signal
import ssl
import typing as t

from gileum import BaseGileum
from hypercorn.asyncio import serve
from hypercorn.config import Config

from . import ASGIApp_t
from .types import (
    Literal,
    Logger_t,
    LoggerClass_t,
)


# Units
BYTES = 1
OCTETS = 1
SECONDS = 1.0


class HypercornGileum(BaseGileum):

    glm_name: Literal["main"] = "main"

    access_log_format: str =\
        '%(h)s %(l)s %(l)s %(t)s "%(r)s" %(s)s %(b)s "%(f)s" "%(a)s"'
    accesslog: t.Union[Logger_t, str, None] = None
    alpn_protocols: t.List[str] = ["h2", "http/1.1"]
    alt_svc_headers: t.List[str] = []
    backlog: int = 100
    bind: t.Union[str, t.List[str]] = ["127.0.0.1:8000"]
    ca_certs: t.Optional[str] = None
    certfile: t.Optional[str] = None
    ciphers: str = "ECDHE+AESGCM"
    debug: bool = False
    dogstatsd_tags: str = ""
    errorlog: t.Union[Logger_t, str, None] = "-"
    graceful_timeout: float = 3 * SECONDS
    group: t.Optional[int] = None
    h11_max_incomplete_size: int = 16 * 1024 * BYTES
    h2_max_concurrent_streams: int = 100
    h2_max_header_list_size: int = 2 ** 16
    h2_max_inbound_frame_size: int = 2 ** 14 * OCTETS
    include_server_header: bool = True
    insecure_bind: t.Union[str, t.List[str]] = []
    keep_alive_timeout: float = 5 * SECONDS
    keyfile: t.Optional[str] = None
    logconfig: t.Optional[str] = None
    logconfig_dict: t.Optional[dict] = None
    logger_class: LoggerClass_t = logging.Logger
    loglevel: str = "INFO"
    max_app_queue_size: int = 10
    quic_bind: t.Union[str, t.List[str]] = []
    pid_path: t.Optional[str] = None
    root_path: str = ""
    server_names: t.List[str] = []
    shutdown_timeout: float = 60 * SECONDS
    ssl_handshake_timeout: float = 60 * SECONDS
    startup_timeout: float = 60 * SECONDS
    statsd_host: t.Optional[str] = None
    statsd_prefix: str = ""
    umask: t.Optional[int] = None
    use_reloader: bool = False
    user: t.Optional[int] = None
    verify_flags: t.Optional[ssl.VerifyFlags] = None
    verify_mode: t.Optional[ssl.VerifyMode] = None
    websocket_max_message_size: int = 16 * 1024 * 1024 * BYTES
    websocket_ping_interval: t.Optional[float] = None
    worker_class: str = "asyncio"
    workers: int = 1


_shutdown_event = asyncio.Event()


def _signal_hadler(*args: t.Any) -> None:
    _shutdown_event.set()


def run_hypercorn(app: ASGIApp_t, glm: HypercornGileum) -> None:
    config = Config.from_mapping(glm.dict(exclude={"glm_name"}))

    # TODO
    #   It's worth re-considering its signal handling.
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGTERM, _signal_hadler)
    loop.run_until_complete(
        serve(app, config, shutdown_trigger=_shutdown_event.wait)
    )



import logging
import sys
import typing as t

if sys.version_info.minor >= 8:
    from typing import Literal
else:
    from typing_extensions import Literal


class _Validater:

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: t.Any) -> t.Any:
        raise NotImplementedError


class Logger_t(_Validater):

    @classmethod
    def validate(cls, v: t.Any) -> logging.Logger:
        if not isinstance(v, logging.Logger):
            raise ValueError(f"{v} is not a logging.Logger.")
        return v


class LoggerClass_t(_Validater):

    @classmethod
    def validate(cls, v: t.Any) -> t.Type[logging.Logger]:
        if not issubclass(v, logging.Logger):
            raise ValueError(f"{v} is not a subclass of logging.Logger.")
        return v



__version__ = "0.0.1"


import typing as t


WSGIApp_t = t.Callable[
    [t.Dict[str, t.Any], t.Callable[[t.Iterable[t.Tuple[str, str]]], None]],
    t.Iterable[bytes],
]
ASGIApp_t = t.Callable[
    [
        t.Dict[str, t.Any],
        t.Callable[[], t.Awaitable[t.Dict[str, t.Any]]],
        t.Callable[[t.Dict[str, t.Any]], t.Awaitable[None]]
    ],
    t.Awaitable[None]
]



import ssl
import typing as t

from gileum import BaseGileum
import gunicorn.glogging
from gunicorn.app.base import BaseApplication
from gunicorn.arbiter import Arbiter
from gunicorn.http.message import Request
from gunicorn.http.wsgi import Response
from gunicorn.workers.base import Worker

from . import WSGIApp_t
from .types import Literal


HookArbiter_t = t.Callable[[Arbiter], None]
HookWorker_t = t.Callable[[Worker], None]
HookArbiterWorker_t = t.Callable[[Arbiter, Worker], None]
HookPreRequest_t = t.Callable[[Worker, Request], None]
HookPostRequest_t = t.Callable[[Worker, Request, t.Dict[str, t.Any], Response], None]
HookNworkersChanged_t = t.Callable[[Arbiter, int, int], None]


class GunicornGileum(BaseGileum):

    # Reference at https://docs.gunicorn.org/en/latest/settings.html

    # Debugging
    reload: bool = False
    reload_engine: Literal["auto", "poll", "inotify"] = "auto"
    reload_extra_files: t.List[str] = []
    spew: bool = False
    check_config: bool = False
    print_config: bool = False

    # Logging
    accesslog: t.Optional[str] = None
    disable_redirect_access_to_syslog: bool = False
    access_log_format: t.Optional[str] = None
    errorlog: str = "-"
    loglevel: Literal[
        "debug",
        "info",
        "warning",
        "error",
        "critical",
    ] = "info"
    capture_output: bool = False
    logger_class: t.Any = gunicorn.glogging.Logger
    logconfig: t.Optional[str] = None
    logconfig_dict: dict = {}
    syslog_addr: str = "udp://localhost:514"
    syslog: bool = False
    syslog_prefix: t.Optional[str] = None
    syslog_facility: str = "user"
    enable_stdio_inheritance: bool = False
    statsd_host: t.Optional[str] = None
    dogstatsd_tags: str = ""
    statsd_prefix: str = ""

    # Process Naming
    proc_name: t.Optional[str] = None
    default_proc_name: str = "gunicorn"

    # SSL
    keyfile: t.Optional[str] = None
    certfile: t.Optional[str] = None
    ssl_version: ssl._SSLMethod = ssl.PROTOCOL_SSLv23
    cert_reqs: ssl.VerifyMode = ssl.VerifyMode.CERT_NONE
    ca_certs: t.Optional[str] = None
    suppress_ragged_eofs: bool = True
    do_handshake_on_connect: bool = False
    ciphers: t.Optional[str] = None

    # Security
    limit_request_line: int = 4094
    limit_request_fields: int = 100
    limit_request_field_size: int = 8190

    # Server Hooks
    on_starting: t.Optional[HookArbiter_t] = None
    on_reload: t.Optional[HookArbiter_t] = None
    when_ready: t.Optional[HookArbiter_t] = None
    pre_fork: t.Optional[HookArbiterWorker_t] = None
    post_fork: t.Optional[HookArbiterWorker_t] = None
    post_worker_init: t.Optional[HookWorker_t] = None
    worker_init: t.Optional[HookWorker_t] = None
    worker_int: t.Optional[HookWorker_t] = None
    worker_abort: t.Optional[HookWorker_t] = None
    pre_exec: t.Optional[HookArbiter_t] = None
    pre_request: t.Optional[HookPreRequest_t] = None
    post_request: t.Optional[HookPostRequest_t] = None
    child_exit: t.Optional[HookArbiterWorker_t] = None
    worker_exit: t.Optional[HookArbiterWorker_t] = None
    nworkers_changed: t.Optional[HookNworkersChanged_t] = None
    on_exit: t.Optional[HookArbiter_t] = None

    # ServerMechanics
    preload_app: bool = False
    sendfile: t.Optional[str] = None
    reuse_port: bool = False
    chdir: t.Optional[str] = None
    daemon: bool = False
    raw_env: t.List[str] = []
    pidfile: t.Optional[str] = None
    worker_tmp_dir: t.Optional[str] = None
    user: t.Optional[int] = None
    group: t.Optional[int] = None
    umask: int = 0
    initgroups: bool = False
    tmp_upload_dir: t.Optional[str] = None
    secure_scheme_headers: t.Optional[t.Dict[str, str]] = None
    forwarded_allow_ips: str = "127.0.0.1"
    pythonpath: t.Optional[str] = None
    paste: t.Optional[str] = None
    proxy_protocol: bool = False
    proxy_allow_ips: str = "127.0.0.1"
    raw_paste_global_conf: t.List[str] = []
    strip_header_spaces: bool = False

    # Server Socket
    bind: t.List[str] = ["127.0.0.1:8000"]
    backlog: int = 2048

    # Worker Processes
    workers: int = 1
    worker_class: Literal[
        "sync",
        "eventlet",
        "gevent",
        "tornado",
        "gthread"
    ] = "sync"
    threads: int = 1
    worker_connections: int = 1000
    max_requests: int = 0
    max_requests_jitter: int = 0
    timeout: int = 30
    graceful_timeout: int = 30
    keepalive: int = 2


class GunicornServer(BaseApplication):

    # Reference at https://docs.gunicorn.org/en/latest/custom.html

    def __init__(self, app, options: t.Dict[str, t.Any] = None):
        self.options = options or {}
        self.app = app
        super().__init__()

    def load_config(self):
        config = {key: value for key, value in self.options.items()
                  if key in self.cfg.settings and value is not None}
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.app


GunicornServer_t = t.TypeVar("GunicornServer_t", bound=GunicornServer)



def run_gunicorn(
    app: WSGIApp_t,
    glm: GunicornGileum,
    server: t.Type[GunicornServer_t] = GunicornServer,
) -> None:
    server(app, glm.dict(exclude={"glm_name"})).run()
